@@ -16,17 +16,15 @@
 
 import sys
 import urllib
-import json
 import time
 import xml.etree.ElementTree as ET
-from tornado.escape import json_decode
-from tornado.httpclient import HTTPClient, HTTPError
-from tornado.httputil import HTTPHeaders
 import sys
 import types
+import requests
+from requests.exceptions import HTTPError
 import pandas as pd
 import numpy as np
-from botocore.client import Config
+from ibm_botocore.client import Config
 import ibm_boto3
 from datetime import datetime
 import pyarrow
@@ -38,10 +36,12 @@ class SQLQuery():
     def __init__(self, api_key, instance_crn, target_cos_url=None, client_info=''):
         self.endpoint_alias_mapping = {
             "us-geo": "s3-api.us-geo.objectstorage.softlayer.net",
+            "us": "s3-api.us-geo.objectstorage.softlayer.net",
             "dal-us-geo": "s3-api.dal-us-geo.objectstorage.softlayer.net",
             "wdc-us-geo": "s3-api.wdc-us-geo.objectstorage.softlayer.net",
             "sjc-us-geo": "s3-api.sjc-us-geo.objectstorage.softlayer.net",
             "eu-geo": "s3.eu-geo.objectstorage.softlayer.net",
+            "eu": "s3.eu-geo.objectstorage.softlayer.net",
             "ams-eu-geo": "s3.ams-eu-geo.objectstorage.softlayer.net",
             "fra-eu-geo": "s3.fra-eu-geo.objectstorage.softlayer.net",
             "mil-eu-geo": "s3.mil-eu-geo.objectstorage.softlayer.net",
@@ -49,6 +49,7 @@ class SQLQuery():
             "us-east": "s3.us-east.objectstorage.softlayer.net",
             "jp-tok": "s3.jp-tok.objectstorage.softlayer.net",
             "ap-geo": "s3.ap-geo.objectstorage.softlayer.net",
+            "ap": "s3.ap-geo.objectstorage.softlayer.net",
             "tok-ap-geo": "s3.tok-ap-geo.objectstorage.softlayer.net",
             "seo-ap-geo": "s3.seo-ap-geo.objectstorage.softlayer.net",
             "hkg-ap-geo": "s3.hkg-ap-geo.objectstorage.softlayer.net",
@@ -58,6 +59,7 @@ class SQLQuery():
             "che01": "s3.che01.objectstorage.softlayer.net",
             "mel01": "s3.mel01.objectstorage.softlayer.net",
             "tor01": "s3.tor01.objectstorage.softlayer.net",
+            "mon01": "s3.mon01.objectstorage.softlayer.net",
             "osl01": "s3.osl01.objectstorage.softlayer.net",
             "sao01": "s3.sao01.objectstorage.softlayer.net",
             "seo01": "s3.seo01.objectstorage.softlayer.net"
@@ -70,13 +72,14 @@ class SQLQuery():
             self.user_agent = 'IBM Cloud SQL Query Python SDK'
         else:
             self.user_agent = client_info
-        self.client = HTTPClient()
-        self.request_headers = HTTPHeaders({'Content-Type': 'application/json'})
-        self.request_headers.add('Accept', 'application/json')
-        self.request_headers.add('User-Agent', self.user_agent)
-        self.request_headers_xml_content = HTTPHeaders({'Content-Type': 'application/x-www-form-urlencoded'})
-        self.request_headers_xml_content.add('Accept', 'application/json')
-        self.request_headers_xml_content.add('User-Agent', self.user_agent)
+
+        self.request_headers = {'Content-Type': 'application/json'}
+        self.request_headers.update({'Accept': 'application/json'})
+        self.request_headers.update({'User-Agent': self.user_agent})
+        self.request_headers_xml_content = {'Content-Type': 'application/x-www-form-urlencoded'}
+        self.request_headers_xml_content.update({'Accept': 'application/json'})
+        self.request_headers_xml_content.update({'User-Agent': self.user_agent})
+
         self.logged_on = False
 
     def logon(self):
@@ -85,45 +88,53 @@ class SQLQuery():
         else:
             data = urllib.urlencode({'grant_type': 'urn:ibm:params:oauth:grant-type:apikey', 'apikey': self.api_key})
 
-        response = self.client.fetch(
+        response = requests.post(
             'https://iam.bluemix.net/identity/token',
-            method='POST',
             headers=self.request_headers_xml_content,
-            validate_cert=False,
-            body=data)
+            data=data)
 
-        if response.code == 200:
+        if response.status_code == 200:
             # print("Authentication successful")
-            bearer_response = json_decode(response.body)
+            bearer_response = response.json()
             self.bearer_token = 'Bearer ' + bearer_response['access_token']
-            self.request_headers = HTTPHeaders({'Content-Type': 'application/json'})
-            self.request_headers.add('Accept', 'application/json')
-            self.request_headers.add('User-Agent', self.user_agent)
-            self.request_headers.add('authorization', self.bearer_token)
+            self.request_headers = {'Content-Type': 'application/json'}
+            self.request_headers.update({'Accept':'application/json'})
+            self.request_headers.update({'User-Agent': self.user_agent})
+            self.request_headers.update({'authorization': self.bearer_token})
             self.logged_on = True
 
         else:
-            print("Authentication failed with http code {}".format(response.code))
+            print("Authentication failed with http code {}".format(response.status_code))
 
-    def submit_sql(self, sql_text):
+    def submit_sql(self, sql_text, pagesize=None):
         if not self.logged_on:
             print("You are not logged on to IBM Cloud")
             return
         sqlData = {'statement': sql_text}
-        if self.target_cos:
+        # If a valid pagesize is specified we need to append the proper PARTITIONED EVERY <num> ROWS clause
+        if pagesize or pagesize==0:
+            if type(pagesize) == int and pagesize>0:
+                if self.target_cos:
+                    sqlData["statement"] += " INTO {}".format(self.target_cos)
+                elif " INTO "  not in sql_text.upper():
+                    raise SyntaxError("Neither resultset_target parameter nor \"INTO\" clause specified.")
+                elif " PARTITIONED " in sql_text.upper():
+                    raise SyntaxError("Must not use PARTITIONED clause when specifying pagesize parameter.")
+                sqlData["statement"] += " PARTITIONED EVERY {} ROWS".format(pagesize)
+            else:
+                raise ValueError('pagesize parameter ({}) is not valid.'.format(pagesize))
+        elif self.target_cos:
             sqlData.update({'resultset_target': self.target_cos})
 
         try:
-            response = self.client.fetch(
+            response = requests.post(
                 "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
-                method='POST',
                 headers=self.request_headers,
-                validate_cert=False,
-                body=json.dumps(sqlData))
+                json=sqlData)
+            resp = response.json()
+            return resp['job_id']
         except HTTPError as e:
-            raise SyntaxError("SQL submission failed: {}".format(json_decode(e.response.body)['errors'][0]['message']))
-
-        return json_decode(response.body)['job_id']
+            raise SyntaxError("SQL submission failed: {}".format(response.json()['errors'][0]['message']))
 
     def wait_for_job(self, jobId):
         if not self.logged_on:
@@ -131,14 +142,13 @@ class SQLQuery():
             return "Not logged on"
 
         while True:
-            response = self.client.fetch(
+            response = requests.get(
                 "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
-                method='GET',
                 headers=self.request_headers,
-                validate_cert=False)
+            )
 
-            if response.code == 200 or response.code == 201:
-                status_response = json_decode(response.body)
+            if response.status_code == 200 or response.status_code == 201:
+                status_response = response.json()
                 jobStatus = status_response['status']
                 if jobStatus == 'completed':
                     # print("Job {} has completed successfully".format(jobId))
@@ -148,7 +158,7 @@ class SQLQuery():
                     print("Job {} has failed".format(jobId))
                     break
             else:
-                print("Job status check failed with http code {}".format(response.code))
+                print("Job status check failed with http code {}".format(response.status_code))
                 break
             time.sleep(2)
         return jobStatus
@@ -156,13 +166,13 @@ class SQLQuery():
     def __iter__(self):
         return 0
 
-    def get_result(self, jobId):
+    def get_result(self, jobId, pagenumber=None):
         if not self.logged_on:
             print("You are not logged on to IBM Cloud")
             return
 
         job_details = self.get_job(jobId)
-        job_status = job_details['status']
+        job_status = job_details.get('status')
         if job_status == 'running':
             raise ValueError('SQL job with jobId {} still running. Come back later.')
         elif job_status != 'completed':
@@ -179,15 +189,14 @@ class SQLQuery():
         if result_format not in ["csv", "parquet"]:
             raise ValueError("Result object format {} currently not supported by get_result().".format(result_format))
 
-        response = self.client.fetch(
+        response = requests.get(
             result_location,
-            method='GET',
             headers=self.request_headers,
-            validate_cert=False)
+        )
 
-        if response.code == 200 or response.code == 201:
+        if response.status_code == 200 or response.status_code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
-            responseBodyXMLroot = ET.fromstring(response.body)
+            responseBodyXMLroot = ET.fromstring(response.text)
             bucket_objects = []
             # Find result objects with data
             for contents in responseBodyXMLroot.findall('s3:Contents', ns):
@@ -197,7 +206,7 @@ class SQLQuery():
                 #print("Job result for {} stored at: {}".format(jobId, result_object))
         else:
             raise ValueError("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
-                                                                                           response.code))
+                                                                                           response.status_code))
 
         cos_client = ibm_boto3.client(service_name='s3',
                                       ibm_api_key_id=self.api_key,
@@ -205,43 +214,61 @@ class SQLQuery():
                                       config=Config(signature_version='oauth'),
                                       endpoint_url='https://' + result_cos_endpoint)
 
-        # Loop over result objects and read and concatenate them into result data frame
-        for bucket_object in bucket_objects:
+        # When pagenumber is specified we only retrieve that page. Otherwise we concatenate all pages to one DF:
+        if pagenumber or pagenumber==0:
+            if " PARTITIONED EVERY " not in job_details['statement'].upper():
+                raise ValueError("pagenumber ({}) specified, but the job was not submitted with pagination option.".format(pagenumber))
+            if type(pagenumber) == int and 0 < pagenumber <= len(bucket_objects):
+                if result_format == "csv":
+                    body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_objects[pagenumber-1])['Body']
+                    if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
+                    result_df = pd.read_csv(body)
+                elif result_format == "parquet":
+                    tmpfile = tempfile.NamedTemporaryFile()
+                    tempfilename = tmpfile.name
+                    tmpfile.close()
+                    cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_objects[pagenumber-1], Filename=tempfilename)
+                    result_df = pd.read_parquet(tempfilename)
+            else:
+                raise ValueError("Invalid pagenumner ({}) specified".format(pagenumber))
+        else:
+            # Loop over result objects and read and concatenate them into result data frame
+            for bucket_object in bucket_objects:
 
-            if result_format == "csv":
-                body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_object)['Body']
-                # add missing __iter__ method, so pandas accepts body as file-like object
-                if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
+                if result_format == "csv":
+                    body = cos_client.get_object(Bucket=result_cos_bucket, Key=bucket_object)['Body']
+                    # add missing __iter__ method, so pandas accepts body as file-like object
+                    if not hasattr(body, "__iter__"): body.__iter__ = types.MethodType(self.__iter__, body)
 
-                partition_df = pd.read_csv(body)
+                    partition_df = pd.read_csv(body)
 
-            elif result_format == "parquet":
-                tmpfile = tempfile.NamedTemporaryFile()
-                tempfilename = tmpfile.name
-                cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_object, Filename=tempfilename)
+                elif result_format == "parquet":
+                    tmpfile = tempfile.NamedTemporaryFile()
+                    tempfilename = tmpfile.name
+                    tmpfile.close()
+                    cos_client.download_file(Bucket=result_cos_bucket, Key=bucket_object, Filename=tempfilename)
 
-                partition_df = pd.read_parquet(tempfilename)
-                tmpfile.close()
+                    partition_df = pd.read_parquet(tempfilename)
 
-            # Add columns from hive style partition naming schema
-            hive_partition_candidates = bucket_object.replace(result_cos_prefix + '/', '').split('/')
-            for hive_partition_candidate in hive_partition_candidates:
-                if hive_partition_candidate.count('=') == 1: # Hive style folder names contain exactly one '='
-                    column = hive_partition_candidate.split('=')
-                    column_name = column[0]
-                    column_value = column[1]
-                    if column_value == '__HIVE_DEFAULT_PARTITION__': # Null value partition
-                        column_value = np.nan
-                    if len(column_name) > 0 and len(column_value) > 0:
-                        partition_df[column_name] = column_value
+                # Add columns from hive style partition naming schema
+                hive_partition_candidates = bucket_object.replace(result_cos_prefix + '/', '').split('/')
+                for hive_partition_candidate in hive_partition_candidates:
+                    if hive_partition_candidate.count('=') == 1: # Hive style folder names contain exactly one '='
+                        column = hive_partition_candidate.split('=')
+                        column_name = column[0]
+                        column_value = column[1]
+                        if column_value == '__HIVE_DEFAULT_PARTITION__': # Null value partition
+                            column_value = np.nan
+                        if len(column_name) > 0 and len(column_value) > 0:
+                            partition_df[column_name] = column_value
+
+                if 'result_df' not in locals():
+                    result_df = partition_df
+                else:
+                    result_df = result_df.append(partition_df)
 
             if 'result_df' not in locals():
-                result_df = partition_df
-            else:
-                result_df = result_df.append(partition_df)
-
-        if 'result_df' not in locals():
-            return None
+                return None
 
         return result_df
 
@@ -262,13 +289,12 @@ class SQLQuery():
 
         fourth_slash = result_location.replace('/', 'X', 3).find('/')
 
-        response = self.client.fetch(
+        response = requests.get(
             result_location[:fourth_slash] + '?prefix=' + result_location[fourth_slash + 1:],
-            method='GET',
             headers=self.request_headers,
-            validate_cert=False)
+        )
 
-        if response.code == 200 or response.code == 201:
+        if response.status_code == 200 or response.status_code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
             responseBodyXMLroot = ET.fromstring(response.body)
             bucket_name = responseBodyXMLroot.find('s3:Name', ns).text
@@ -284,7 +310,7 @@ class SQLQuery():
                 return
         else:
             print("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
-                                                                                           response.code))
+                                                                                           response.status_code))
             return
 
         result_objects_df = pd.DataFrame(columns=['ObjectURL', 'Size'])
@@ -310,13 +336,12 @@ class SQLQuery():
 
         fourth_slash = result_location.replace('/', 'X', 3).find('/')
 
-        response = self.client.fetch(
+        response = requests.get(
             result_location[:fourth_slash] + '?prefix=' + result_location[fourth_slash + 1:],
-            method='GET',
             headers=self.request_headers,
-            validate_cert=False)
+            )
 
-        if response.code == 200 or response.code == 201:
+        if response.status_code == 200 or response.status_code == 201:
             ns = {'s3': 'http://s3.amazonaws.com/doc/2006-03-01/'}
             responseBodyXMLroot = ET.fromstring(response.body)
             bucket_name = responseBodyXMLroot.find('s3:Name', ns).text
@@ -330,7 +355,7 @@ class SQLQuery():
                 return
         else:
             print("Result object listing for job {} at {} failed with http code {}".format(jobId, result_location,
-                                                                                           response.code))
+                                                                                           response.status_code))
             return
 
         cos_client = ibm_boto3.client(service_name='s3',
@@ -353,43 +378,40 @@ class SQLQuery():
             return
 
         try:
-            response = self.client.fetch(
+            response = requests.get(
                 "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(jobId, self.instance_crn),
-                method='GET',
                 headers=self.request_headers,
-                validate_cert=False)
+            )
         except HTTPError as e:
-            if e.response.code == 400:
+            if e.response.status_code == 400:
                 raise ValueError("SQL jobId {} unknown".format(jobId))
             else:
                 raise e
 
-        return json_decode(response.body)
+        return response.json()
 
     def get_jobs(self):
         if not self.logged_on:
             print("You are not logged on to IBM Cloud")
             return
 
-        response = self.client.fetch(
+        response = requests.get(
             "https://sql-api.ng.bluemix.net/v2/sql_jobs?instance_crn={}".format(self.instance_crn),
-            method='GET',
             headers=self.request_headers,
-            validate_cert=False)
-        if response.code == 200 or response.code == 201:
-            job_list = json_decode(response.body)
+            )
+        if response.status_code == 200 or response.status_code == 201:
+            job_list = response.json()
             job_list_df = pd.DataFrame(columns=['job_id', 'status', 'user_id', 'statement', 'resultset_location',
                                                 'submit_time', 'end_time', 'rows_read', 'rows_returned', 'bytes_read',
                                                 'error', 'error_message'])
             for job in job_list['jobs']:
-                response = self.client.fetch(
+                response = requests.get(
                     "https://sql-api.ng.bluemix.net/v2/sql_jobs/{}?instance_crn={}".format(job['job_id'],
                                                                                                 self.instance_crn),
-                    method='GET',
                     headers=self.request_headers,
-                    validate_cert=False)
-                if response.code == 200 or response.code == 201:
-                    job_details = json_decode(response.body)
+                    )
+                if response.status_code == 200 or response.status_code == 201:
+                    job_details = response.json()
                     error = None
                     error_message = None
                     rows_read = None
@@ -423,16 +445,16 @@ class SQLQuery():
                                                        }], ignore_index=True)
                 else:
                     print("Job details retrieval for jobId {} failed with http code {}".format(job['job_id'],
-                                                                                               response.code))
+                                                                                               response.status_code))
                     break
         else:
-            print("Job list retrieval failed with http code {}".format(response.code))
+            print("Job list retrieval failed with http code {}".format(response.status_code))
         return job_list_df
 
-    def run_sql(self, sql_text):
+    def run_sql(self, sql_text, pagesize=None):
         self.logon()
         try:
-            jobId = self.submit_sql(sql_text)
+            jobId = self.submit_sql(sql_text, pagesize)
         except SyntaxError as e:
             return "SQL job submission failed. {}".format(str(e))
         if self.wait_for_job(jobId) == 'failed':
@@ -570,4 +592,4 @@ class SQLQuery():
             print("Exported {} new jobs".format(new_jobs_df['job_id'].count()))
             tmpfile.close()
         else:
-            print("No new jobs to export")
\ No newline at end of file
+            print("No new jobs to export")
